# ExtendedEvents release 3 by David Bowland
# ./addons/eventscripts/extendedevents/extendedevents.py

# >>> To configure this addon please see extendedevents.cfg <<<

"""
Creates and fires new custom events: addon_action, map_end, player_activeweapon, player_air, player_bombzone, player_buyzone, player_drop, player_duck, player_flag, player_health, player_land, player_money, player_move, player_nightvision, player_rescuezone, player_unduck. Requires only ES 2.0+
"""


import es
import os.path

import psyco
psyco.full()


"""
New events provided by ExtendedEvents:

addon_action
map_end
player_activeweapon
player_air
player_bombzone
player_buyzone
player_drop
player_duck
player_flag
player_health
player_land
player_money
player_move
player_nightvision
player_rescuezone
player_unduck
"""


info = es.AddonInfo()
info.name     = 'ExtendedEvents'
info.version  = '3'
info.url      = 'http://addons.eventscripts.com/addons/view/extendedevents'
info.basename = 'extendedevents'
info.author   = 'SuperDave'


int_ticks        = 0
int_added_prefix = 0
dict_players     = {}
dict_addons      = {}
bool_cstrike     = es.getgame() in ('Counter-Strike: Source', 'cstrike')
bool_map_ended   = False


dict_events  = {'addon_action':{'name':'string', 'action':'string', 'type':'string'}, 'map_end':{}, 'player_activeweapon':{'userid':'short', 'weapon':'string', 'old_weapon':'string'}, 'player_air':{'userid':'short'}, 'player_bombzone':{'userid':'short', 'action':'string'}, 'player_buyzone':{'userid':'short', 'action':'string'}, 'player_drop':{'userid':'short', 'weapon':'string', 'index':'short'}, 'player_duck':{'userid':'short'}, 'player_flag':{'userid':'short', 'flag':'short', 'state':'string'}, 'player_health':{'userid':'short', 'old_health':'short', 'change_health':'short'}, 'player_land':{'userid':'short'}, 'player_money':{'userid':'short', 'amount':'short', 'old_amount':'short', 'change_amount':'short'}, 'player_move':{'userid':'short', 'x':'short', 'y':'short', 'z':'short'}, 'player_nightvision':{'userid':'short', 'state':'string'}, 'player_rescuezone':{'userid':'short', 'action':'string'}, 'player_unduck':{'userid':'short'}}
dict_options = {'extendedevents_tickstoupdate':[16, 'Number of ticks between ExtendedEvents updates'], 'extendedevents_noisy':[0, '0 = ExtendedEvents will not fire noisy events, 1 = ExtendedEvents will fire noisy events\nCurrently only player_move is a noisy event']}


def load():
   """
   Checks to make sure critical server variables are created by the config
   Calls player_spawn for each player
   Registers a client command filter to client_filter
   Registers a tick listener to do_tick
   """
   global int_added_prefix
   global dict_addons

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('extendedevents') + '/extendedevents.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/extendedevents/extendedevents.cfg')
   else:
      es.dbgmsg(0, 'ExtendedEvents: Unable to load extendedevents.cfg! Please ensure it is in the ./extendedevents/ directory.')
   load_events()

   dict_player_list = es.createplayerlist()
   dict_entity_list = es.createentitylist()
   list_entity_list = filter(lambda x: dict_entity_list[x]['classname'].startswith('weapon_'), dict_entity_list)
   for int_userid in dict_player_list:
      player_spawn({'userid':str(int_userid), 'es_userweapon':dict_player_list[int_userid]['weapon']})
      int_player_handle = es.getplayerhandle(int_userid)
      dict_players[int_userid]['weapon_indexes'][:] = filter(lambda x: es.getindexprop(x, 'CBaseEntity.m_hOwnerEntity') == int_player_handle, list_entity_list)

   dict_addons = get_script_list()
   fire_event('addon_action', {'name':{'type':'setstring', 'value':'extendedevents'}, 'action':{'type':'setstring', 'value':'LOAD'}, 'type':{'type':'setstring', 'value':'PY'}})

   es.addons.registerClientCommandFilter(client_filter)
   es.addons.registerTickListener(execute_tick)


def es_map_start(event_var):
   """
   Clears dict_players of players so they are recreated for the new map
   Calls load_events
   """
   global bool_map_ended

   dict_players.clear()

   load_events()

   bool_map_ended = False


def item_pickup(event_var):
   """Adds the picked up weapon to the list of the player's weapons"""
   global dict_players

   int_userid = int(event_var['userid'])
   if dict_players.has_key(int_userid):
      dict_current_player = dict_players[int_userid]
      int_player_handle   = es.getplayerhandle(int_userid)

      for int_index in filter(lambda x: es.getindexprop(x, 'CBaseEntity.m_hOwnerEntity') == int_player_handle and x not in dict_current_player['weapon_indexes'], es.createentitylist('weapon_' + event_var['item'])):
         dict_current_player['weapon_indexes'].append(int_index)


def player_spawn(event_var):
   """
   Creates the spawning player in dict_players if the player doesn't exist
   Resets the variables on the spawning player to prevent all events from firing at once
   """
   global dict_players

   int_userid = int(event_var['userid'])
   if not dict_players.has_key(int_userid):
      dict_players[int_userid] = {'weapon_indexes':[]}

   dict_current_player = dict_players[int_userid]
   dict_current_player['player_activeweapon'] = event_var['es_userweapon']
   dict_current_player['player_bombzone']     = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInBombZone') % 2) if bool_cstrike else 0
   dict_current_player['player_buyzone']      = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInBuyZone') % 2) if bool_cstrike else 0
   dict_current_player['player_duck']         = es.getplayerprop(int_userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2
   dict_current_player['player_health']       = es.getplayerprop(int_userid, 'CBasePlayer.m_iHealth')
   dict_current_player['player_money']        = (es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')) if bool_cstrike else 0
   dict_current_player['player_move']         = es.getplayerlocation(int_userid)
   dict_current_player['player_nightvision']  = (es.getplayerprop(int_userid, 'CCSPlayer.m_bNightVisionOn') % 2) if bool_cstrike else 0
   dict_current_player['player_rescuezone']   = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInHostageRescueZone') % 2) if bool_cstrike else 0
   dict_current_player['player_flags']        = es.getplayerprop(int_userid, 'CBasePlayer.m_fFlags')

   int_player_handle = es.getplayerhandle(int_userid)
   dict_entity_list  = es.createentitylist()
   for int_index in filter(lambda x: (es.getindexprop(x, 'CBaseEntity.m_hOwnerEntity') <> int_player_handle) if dict_entity_list.has_key(x) else False, dict_current_player['weapon_indexes']):
      dict_current_player['weapon_indexes'].remove(int_index)


def player_death(event_var):
   """Clears the list of weapons held by the dead player"""
   global dict_players

   int_userid = int(event_var['userid'])
   if dict_players.has_key(int_userid):
      dict_players[int_userid]['weapon_indexes'][:] = []


def player_disconnect(event_var):
   """Removes the disconnecting player from dict_players"""
   global dict_players

   int_userid = int(event_var['userid'])
   if dict_players.has_key(int_userid):
      del dict_players[int_userid]


def unload():
   """
   Unregisters the tick listener to do_tick
   Removes the client command filter
   """
   es.addons.unregisterTickListener(execute_tick)

   fire_event('addon_action', {'name':{'type':'setstring', 'value':'extendedevents'}, 'action':{'type':'setstring', 'value':'UNLOAD'}, 'type':{'type':'setstring', 'value':'PY'}})

   es.addons.unregisterClientCommandFilter(client_filter)

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


def execute_tick():
   """
   Updates tick count
   addon_action
   Begins loop through all players creating players in dict_players as necessary
      player_activeweapon
      player_flag, player_air, and player_land
      player_bombzone
      player_buyzone
      player_duck and player_unduck
      player_health
      player_money
      player_nightvision
      player_rescuezone
      player_move * "Noisy"
   map_end
   """
   global int_ticks
   global dict_players
   global dict_addons
   global bool_map_ended

   int_ticks += 1
   if int_ticks < es.ServerVar('extendedevents_tickstoupdate'): return

   int_ticks = 0

   dict_new_addons = get_script_list()
   list_new_addons = dict_new_addons.keys()
   for str_addon in dict_addons:
      dict_current_addon = dict_addons[str_addon]
      if dict_new_addons.has_key(str_addon):
         if dict_current_addon['status'] <> dict_new_addons[str_addon]['status']:
            fire_event('addon_action', {'name':{'type':'setstring', 'value':str_addon}, 'action':{'type':'setstring', 'value':'DISABLE' if dict_current_addon['status'] == 'enabled' else 'ENABLE'}, 'type':{'type':'setstring', 'value':dict_current_addon['type'].upper()}})
         list_new_addons.remove(str_addon)

      else:
         fire_event('addon_action', {'name':{'type':'setstring', 'value':str_addon}, 'action':{'type':'setstring', 'value':'UNLOAD'}, 'type':{'type':'setstring', 'value':dict_current_addon['type'].upper()}})

   for str_addon in list_new_addons:
      if not dict_addons.has_key(str_addon):
         fire_event('addon_action', {'name':{'type':'setstring', 'value':str_addon}, 'action':{'type':'setstring', 'value':'LOAD'}, 'type':{'type':'setstring', 'value':dict_new_addons[str_addon]['type'].upper()}})
   dict_addons = dict_new_addons

   dict_player_list = es.createplayerlist()
   bool_map_end     = False if bool_map_ended else bool(dict_player_list)
   for int_userid in dict_players:
      dict_current_player      = dict_players[int_userid]
      dict_current_list_player = dict_player_list[int_userid]

      str_new_value = dict_current_list_player['weapon']
      if dict_current_player['player_activeweapon'] <> str_new_value:
         fire_event('player_activeweapon', {'userid':{'type':'setint', 'value':int_userid}, 'weapon':{'type':'setstring', 'value':str_new_value}, 'old_weapon':{'type':'setstring', 'value':dict_current_player['player_activeweapon']}})
         dict_current_player['player_activeweapon'] = str_new_value

      int_new_value = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInBombZone') % 2) if bool_cstrike else 0
      if dict_current_player['player_bombzone'] <> int_new_value:
         fire_event('player_bombzone', {'userid':{'type':'setint', 'value':int_userid}, 'action':{'type':'setstring', 'value':'ENTER' if int_new_value else 'EXIT'}})
         dict_current_player['player_bombzone'] = int_new_value

      int_new_value = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInBuyZone') % 2) if bool_cstrike else 0
      if dict_current_player['player_buyzone'] <> int_new_value:
         fire_event('player_buyzone', {'userid':{'type':'setint', 'value':int_userid}, 'action':{'type':'setstring', 'value':'ENTER' if int_new_value else 'EXIT'}})
         dict_current_player['player_buyzone'] = int_new_value

      int_new_value = es.getplayerprop(int_userid, 'CBasePlayer.localdata.m_Local.m_bDucked') % 2
      if dict_current_player['player_duck'] <> int_new_value:
         fire_event('player_duck' if int_new_value else 'player_unduck', {'userid':{'type':'setint', 'value':int_userid}})
         dict_current_player['player_duck'] = int_new_value

      float_new_flags = es.getplayerprop(int_userid, 'CBasePlayer.m_fFlags')
      for int_x in range(32):
         int_new_state = float_new_flags >> int_x & 1
         if int_new_state <> dict_current_player['player_flags'] >> int_x & 1:
            if not int_x:
               fire_event('player_land' if int_new_state else 'player_air', {'userid':{'type':'setint', 'value':int_userid}})
            fire_event('player_flag', {'userid':{'type':'setint', 'value':int_userid}, 'flag':{'type':'setint', 'value':int_x}, 'state':{'type':'setstring', 'value':'ON' if int_new_state else 'OFF'}})
         if int_x == 5:
            if not int_new_state:
               bool_map_end = False
      dict_current_player['player_flags'] = float_new_flags

      int_new_health    = dict_current_list_player['health']
      int_health_change = int_new_health - dict_current_player['player_health']
      if int_health_change:
         fire_event('player_health', {'userid':{'type':'setint', 'value':int_userid}, 'old_health':{'type':'setint', 'value':dict_current_player['player_health']}, 'change_health':{'type':'setint', 'value':int_health_change}})
         dict_current_player['player_health'] = int_new_health

      int_new_cash    = (es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')) if bool_cstrike else 0
      int_cash_change = int_new_cash - dict_current_player['player_money']
      if int_cash_change:
         fire_event('player_money', {'userid':{'type':'setint', 'value':int_userid}, 'amount':{'type':'setint', 'value':int_new_cash}, 'old_amount':{'type':'setint', 'value':dict_current_player['player_money']}, 'change_amount':{'type':'setint', 'value':int_cash_change}})
         dict_current_player['player_money'] = int_new_cash

      int_new_value = (es.getplayerprop(int_userid, 'CCSPlayer.m_bNightVisionOn') % 2) if bool_cstrike else 0
      if dict_current_player['player_nightvision'] <> int_new_value:
         fire_event('player_nightvision', {'userid':{'type':'setint', 'value':int_userid}, 'state':{'type':'setstring', 'value':'ON' if int_new_value else 'OFF'}})
         dict_current_player['player_nightvision'] = int_new_value

      int_new_value = (es.getplayerprop(int_userid, 'CCSPlayer.m_bInHostageRescueZone') % 2) if bool_cstrike else 0
      if dict_current_player['player_rescuezone'] <> int_new_value:
         fire_event('player_rescuezone', {'userid':{'type':'setint', 'value':int_userid}, 'action':{'type':'setstring', 'value':'ENTER' if int_new_value else 'EXIT'}})
         dict_current_player['player_rescuezone'] = int_new_value

      if int(es.ServerVar('extendedevents_noisy')):
         list_player_coords = es.getplayerlocation(int_userid)
         int_x = abs(list_player_coords[0] - dict_current_player['player_move'][0])
         int_y = abs(list_player_coords[1] - dict_current_player['player_move'][1])
         int_z = abs(list_player_coords[2] - dict_current_player['player_move'][2])
         if int_x or int_y or int_z:
            fire_event('player_move', {'userid':{'type':'setint', 'value':int_userid}, 'x':{'type':'setfloat', 'value':int_x}, 'y':{'type':'setfloat', 'value':int_y}, 'z':{'type':'setfloat', 'value':int_z}})
            dict_players[int_userid]['player_move'] = list(list_player_coords)

   if bool_map_end:
      fire_event('map_end')
      bool_map_ended = True


def fire_event(str_event, dict_event_vars={}):
   """Fires the event str_event with dict_event_vars event_vars"""
   es.dbgmsg(1, 'ExtendedEvents: Firing event %s' % str_event)

   es.event('initialize', str_event)

   for str_event_var in dict_event_vars:
      dict_current_var = dict_event_vars[str_event_var]
      es.event(dict_current_var['type'], str_event, str_event_var, dict_current_var['value'])

   es.event('fire', str_event)


def client_filter(int_userid, list_args):
   """Fires player_drop when the player executes the command "drop" """
   global dict_players

   if list_args[0] == 'drop':
      if dict_players.has_key(int_userid):
         dict_current_player = dict_players[int_userid]
         str_weapon          = dict_current_player['player_activeweapon']
         int_weapon_index    = -1
         int_player_handle   = es.getplayerhandle(int_userid)
         dict_entity_list    = es.createentitylist(str_weapon)

         for int_index in dict_current_player['weapon_indexes']:
            if dict_entity_list.has_key(int_index):
               int_weapon_index = int_index
               break

         if int_weapon_index <> -1:
            if es.getindexprop(int_weapon_index, 'CBaseEntity.m_hOwnerEntity') == -1:
               dict_current_player['weapon_indexes'].remove(int_weapon_index)
               fire_event('player_drop', {'userid':{'type':'setint', 'value':int_userid}, 'weapon':{'type':'setstring', 'value':str_weapon}, 'index':{'type':'setint', 'value':int_weapon_index}})

      else:
         es.dbgmsg(0, 'ExtendedEvents: WARNING: player_drop unable to recognize user %s!' % int_userid)
         fire_event('player_drop', {'userid':{'type':'setint', 'value':int_userid}, 'weapon':{'type':'setstring', 'value':'none'}, 'index':{'type':'setint', 'value':-1}})

   return True


def get_script_list():
   """
   Appends the Python addons to the end of the createscriptlist dictionary (only if Python addons such as ExtendedEvents aren't already included)
   """
   dict_script_list = es.createscriptlist()

   if not dict_script_list.has_key('extendedevents'):
      for module_addon in es.addons.getAddonList():
         str_name = module_addon.__name__
         dict_script_list[str_name[0:str_name.rfind('.')].replace('.', '/')] = {'status':'enabled', 'type':'py'}

   return dict_script_list


def load_events():
   """Loads the .res file for the events"""
   if os.path.isfile(es.getAddonPath('extendedevents') + '/extendedevents.res'):
      es.loadevents('declare', 'addons/eventscripts/extendedevents/extendedevents.res')

   else:
      if not os.path.isfile(es.getAddonPath('extendedevents') + 'es_extendedevents_events_db.txt'):
         es.keygroupcreate('extendedevents_events')
         for str_event in dict_events:
            es.keycreate('extendedevents_events', str_event)
            for str_event_var in dict_events[str_event]:
               es.keysetvalue('extendedevents_events', str_event, str_event_var, dict_events[str_event][str_event_var])
         es.keygroupsave('extendedevents_events', '|extendedevents')
         es.keygroupdelete('extendedevents_events')
      es.loadevents('declare', 'addons/eventscripts/extendedevents/es_extendedevents_events_db.txt')
      es.dbgmsg(0, 'ExtendedEvents: Unable to load extendedevents.res! Please ensure it is in the ./extendedevents/ directory.')